# -*- coding: utf-8 -*-
# ASternGrossesFeld_mitBesucht.py
# In dieser Version wird beruecksichtigt, dass ein Feld,
# das bereits mit kuerzerer aktueller Weglaenge erreicht wurde,
# keine bessere Loesung erbringen kann.
# => Besuchtliste mit jeweils aktueller Laenge des Wegs speichern
# Die Besuchtliste wird daher als Dictionary gefuehrt
# => alle Knoten muessen Tupel sein

# Ausserdem wird die Liste wegen der schnelleren Zugriffe
# jeweils am Kopf verlaengert.

### Graph ist Vorgabe einer Schuelergruppe

# ----- A*-Graph -----
# Es werden in einem Dictionary die gesperrten Felder angegeben.
gesperrt={
    1:[2,6,14,18],
    2:[3,9,14,18],
    3:[2,6,8,14],
    4:[3,6,9,14,17,18,19,20],
    5:[14],
    6:[3,5,6,14],
    7:[3,7,10,11,12,14],
    8:[3,4,5,6,11],
    9:[13,14,15,16,17,18,19,20],
    10:[2,5,10],
    11:[2,6,9,13,14,15,16,17,18,19,20],
    12:[2,7,8],
    13:[],
    14:[2,20],
    15:[5,8,10,11,12,13,15,18],
    16:[5,8,10,15,18],
    17:[5,6,7,8,10,13,15,16,17,18],
    18:[5,8,10,13,15,18],
    19:[5,8,10,11,12,13,15,18],
    20:[2,20]
}
# zugehoerige Konstante:
kantenlaenge=20

# ----- Alternative als Liste -----
##gesperrt=[[], ## dummy um den Startindex 1 zu erzielen
##          [2,6,14,18],
##          [3,9,14,18],
##          [2,6,8,14],
##          [3,6,9,14,17,18,19,20],
##          [14],
##          [3,5,6,14],
##          [3,7,10,11,12,14],
##          [3,4,5,6,11],
##          [13,14,15,16,17,18,19,20],
##          [2,5,10],
##          [2,6,9,13,14,15,16,17,18,19,20],
##          [2,7,8],
##          [],
##          [2,20],
##          [5,8,10,11,12,13,15,18],
##          [5,8,10,15,18],
##          [5,6,7,8,10,13,15,16,17,18],
##          [5,8,10,13,15,18],
##          [5,8,10,11,12,13,15,18],
##          [2,20]
##          ]

# ----- unzulaessig -----
# gibt von einem Feld an, ob es zulässig ist.
def zulaessig(feld):
    for koordinate in feld:
        if koordinate<1: return False
        if koordinate>kantenlaenge: return False
    if feld[1] in gesperrt[feld[0]]: return False
    return True

# ----- nachfolger -----
# bestimmt die zulässigen Nachfolgefelder eines Felds
def nachfolger(feld):
    ergebnis=[]
    for fall in [(-1,0),(1,0),(0,-1),(0,1)]:
        if zulaessig((feld[0]+fall[0],feld[1]+fall[1])):
            ergebnis.append((feld[0]+fall[0],feld[1]+fall[1]))
    return ergebnis

# ----- restkosten -----
# Bestimmt zu einem Feld seine geschätzten Restkosten
# nach der Manhattan-Distanz.
def restkosten(feld):
    return abs(feld[0]-ziel[0])+abs(feld[1]-ziel[1])

# ----- kosten -----
# Bestimmt zu einem Weg seine Kosten.
def kosten(weg):
    return len(weg)-1

# ----- gesamtKosten -----
# bestimmt die Gesamtkosten.
def gesamtKosten(weg):
    return kosten(weg)+restkosten(weg[0])

# ----- prioWS -----
# Funktionen zur Verwaltung der Prioritaetswarteschlange

# ----- kuerzer -----
# Hier muss das Vergleichsprädikat kuerzer neu definiert werden.
def kuerzer(weg1, weg2):
    return gesamtKosten(weg1)<gesamtKosten(weg2)

# ----- expandiere -----
# Die Expansionsfunktion neu definiert werden. Neben dem Filtern auf
# Zyklen muessen die laengeren aktuellen Wege zu Knoten der Besuchtliste
# gefiltert werden.
def expandiere(weg, besucht):
    aktuell=weg[0]
    neueKnoten=nachfolger(aktuell)
    gefiltert=[]
    for knoten in neueKnoten:
        if not knoten in weg:
            gefiltert.append(knoten)
    neueWege=[]
    for i in range(len(gefiltert)):
        neueWege.append([gefiltert[i]]+weg)
    return neueWege
    

# ----- ordneEin -----
# Die Funktion ordnet einen einzelnen Weg ein.
# Zunaechst wird bei Wegen, die zu einem besuchten Knoten fuehren,
# geprueft, ob sie einen Weg mit kuerzerer aktueller Laenge bilden
# als der vorhandene in den sortierten
def ordneEin(weg, sortierte):
    index=-1 # Annahme neuer Knoten
    ##input(str(weg)+"\n"+str(sortierte))
    for i in range(len(sortierte)):
        if weg[0]==sortierte[i][0]: # selber aktueller Knoten
            index=i
            if len(weg)>=len(sortierte[i]): # altes bleibt
                return sortierte
    if index>-1:
        sortierte=sortierte[:index]+sortierte[(index+1):]
    for i in range(len(sortierte)): 
        if kuerzer(weg, sortierte[i]):
            sortierte[i:i]=[weg]
            return sortierte
    sortierte.append(weg) # am Ende anhaengen
    return sortierte

# ----- ordneAlleEin -----
# Die Funktion ordnet eine Liste von wegn ein.
# [sortiert durch Einfügen]
def ordneAlleEin(wege, sortierte):
    for weg in wege:
        sortierte=ordneEin(weg, sortierte)
    return sortierte

# ----- aktualisiereBesucht -----
# fuegt die neuen ersten Knoten der besuchtListe hinzu
def aktualisiereBesucht(neue,besucht):
    for neu in neue:
        besucht.update({neu[0]:len(neu)})
    return besucht

# ----- expansionsSchritt -----
# nimmt das erste Element aus der Prioritätswarteschlange
# expandiert es und fügt die entstandenen Wege ein.
#zu jedem aktuellen Knoten darf es in der PrioWS bisher nur einen Weg geben,
#verwendet wird der neue, wenn der alte eine kuerzere aktuelle Laenge hat

def expansionsSchritt(prioWS, ziel, besucht):
    wegZumZiel=None
    while wegZumZiel==None:
        if len(prioWS)==0: return False
        neue=expandiere(prioWS[0],besucht)
        wegZumZiel=amZiel(prioWS, ziel)
        if wegZumZiel: return wegZumZiel
##        if input('\nprioWS'+str(prioWS)
##                 +'\nbesucht'+str(besucht)
##                 +'\nziel'+str(ziel))=='q': return None
        prioWS=ordneAlleEin(neue, prioWS[1:])
        besucht=aktualisiereBesucht(neue,besucht)

def amZiel(neue, ziel):
    for weg in neue:
        if ziel in weg:
            return weg
    return None


# ----- aStern -----
def aStern(von, nach):
    print('Start',von,'Ziel:',nach)
    global ziel
    ziel=nach
    return expansionsSchritt([[von]], ziel, {von:0})

##print( aStern((3,3),(8,10)) )


### ===== zeigeFeldUndWeg =================================
def zeigeFeldUndWeg(breite,hoehe,weg):
    print("\nFeld und Weg\n")
    for i in range(1,breite+1):
        zeile=""
        for j in range(1,hoehe+1):
            if j in gesperrt[i]: zeile+="X "
            elif (i,j) in weg: zeile+="o "
            else: zeile+=". "
        print(zeile)
    print('Gesamtkosten:',gesamtKosten(weg))
    print()

##zeigeFeldUndWeg(20,20,aStern((3,3),(9,10)))
zeigeFeldUndWeg(20,20,aStern((3,3),(17,12)))
